/*
 * Copyright (c) 1999, 2001, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2017-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package javax.net.ssl;

import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;

import ej.annotation.Nullable;

/**
 * This class defines the <i>Service Provider Interface</i> (<b>SPI</b>) for the <code>KeyManagerFactory</code> class.
 *
 * <p>
 * All the abstract methods in this class must be implemented by each cryptographic service provider who wishes to
 * supply the implementation of a particular key manager factory.
 *
 * @since 1.4
 * @see KeyManagerFactory
 * @see KeyManager
 */
public abstract class KeyManagerFactorySpi {
	/**
	 * Initializes this factory with a source of key material.
	 *
	 * @param ks
	 *            the key store or null
	 * @param password
	 *            the password for recovering keys
	 * @throws KeyStoreException
	 *             if this operation fails
	 * @throws NoSuchAlgorithmException
	 *             if the specified algorithm is not available from the specified provider.
	 * @throws UnrecoverableKeyException
	 *             if the key cannot be recovered
	 * @see KeyManagerFactory#init(KeyStore, char[])
	 */
	protected abstract void engineInit(@Nullable KeyStore ks, char[] password)
			throws KeyStoreException, NoSuchAlgorithmException, UnrecoverableKeyException;

	/**
	 * Returns one key manager for each type of key material.
	 *
	 * @return the key managers
	 * @throws IllegalStateException
	 *             if the KeyManagerFactorySpi is not initialized
	 */
	protected abstract KeyManager[] engineGetKeyManagers();
}
